/*
 * plcsrvthread.cpp
 *
 *	2009/07/16 - SF: first release  
 *
 *
 */

#include "plcsrvthread.h"
#include <iomanip>      //used for debug dump of db
#include <netdb.h>

namespace PLCRack3_ns{


/* 
 *	plcsrvthread::run()
 *	Run  
 */		
void plcsrvthread::run(void *)
{
	DEBUG_STREAM << "plcsrvthread::run() entering..." << endl;
	int pausemicro = 3000000;
	server_started=0;
	push_counter = 0;
	allow_push = true;


	buf = new char[2*BUF_LEN];
	while (!t_PLCRack3->abortflag) {
		if(server_started==0)
		{
			Srv_Prepare();
		}
		else
		{
			t_PLCRack3->set_state(Tango::FAULT);
			t_PLCRack3->set_status("srv not connected");
		}
		
		usleep(pausemicro);	
		DEBUG_STREAM << "plcsrvthread::run() trying to reconnect" << endl;
	}
	delete [] buf;
}




void plcsrvthread::Srv_Prepare(void)
{
	DEBUG_STREAM << "plcsrvthread::"<<__func__<<": entering..." << endl;
	char *save_buf_ptr = &buf[0];;
	timeval t_now;
	gettimeofday(&last_read, NULL);
	while (!t_PLCRack3->abortflag)
	{
		bool connection_ok = false;

		server_started=1; //fake
		Tango::DevState dp_state=Tango::FAULT;
		try
		{
			dp_state = t_PLCRack3->socket_get_state();
			if(dp_state == Tango::ON)
				connection_ok = true;
		}
		catch(Tango::DevFailed &e)
		{
			INFO_STREAM << __func__ << ": error reading socket state="<<e.errors[0].desc;
			sleep(1);
		}

		if(connection_ok)
		{//connection established
			DEBUG_STREAM<<"plcsrvthread::Srv_Prepare() connection ok"<<endl;
			errcount =0;
			first_mesg = true;
			cnt = 0;
			buf = save_buf_ptr;
			memset (buf,0,2*BUF_LEN);

			do
			{
				gettimeofday(&t_now, NULL);
				double diff_s = t_now.tv_sec - last_read.tv_sec;
				if(diff_s > 30)
				{
					t_PLCRack3->set_state(Tango::ALARM);
					t_PLCRack3->set_status("Timeout waiting PLC data");
				}

				while(cnt > 0)
				{
					header_t *head;
					head = reinterpret_cast<header_t *>(&buf[0]);
					uint16_t btype = head->db_number;
					btype = ntohs(btype);
					if(btype == 0)
					{
						cnt--;
						buf++;
						DEBUG_STREAM << "plcsrvthread::run() looking for header, now " << cnt << " in buffer"<<endl;
						if(cnt == 0)
						{
							buf = save_buf_ptr;
							memset (buf,0,2*BUF_LEN);
							INFO_STREAM << __func__ << ": " << cnt << " unknown bytes in buffer, type="<<btype<<", emptying";
						}
					}
					else
					{
						break;
					}
				}

				//DEBUG_STREAM << "plcsrvthread::Srv_Prepare()going to read from sock" << endl;
				try
				{
					Tango::DevLong count=t_PLCRack3->input_length();
					if(count > 0)
					{
						if(count+cnt > BUF_LEN)
							count = BUF_LEN-cnt;
						vector<unsigned char> dtr;
						t_PLCRack3->read(count, dtr);
						for (unsigned int i=0; i<dtr.size(); ++i)
						{
							buf[i] = dtr[i];
						}
						gettimeofday(&last_read, NULL);
						INFO_STREAM << __func__ << ": read " << count << " now " << cnt+count << " in buffer" << endl;
					}
					else
					{
						timespec time_to_sleep;
						time_to_sleep.tv_sec=0;
						//time_to_sleep.tv_nsec= 0.5 * 1000000L;//0.5ms
						//time_to_sleep.tv_nsec= 10 * 1000000L;//100ms
						time_to_sleep.tv_nsec= 100 * 1000000L;//100ms
						//time_to_sleep.tv_nsec= 500 * 1000000L;//500ms
						nanosleep(&time_to_sleep, NULL);
					}
					cnt += count;
					if(t_PLCRack3->get_state() != Tango::RUNNING)
					{
						t_PLCRack3->set_status("PLC OK");
						t_PLCRack3->set_state(Tango::RUNNING);
					}
				}
				catch(Tango::DevFailed &e)
				{
					stringstream tmp;
					tmp << "Error reading " << e.errors[0].desc;
					INFO_STREAM << __func__ << ": " << tmp.str();
					cnt = -1;
					t_PLCRack3->set_state(Tango::FAULT);
					t_PLCRack3->set_status(tmp.str());
				}

				if (cnt < 0 ) 
				{
					break; 
				}
				else if (cnt == 0)
				{
					//DEBUG_STREAM<<"plcsrvthread::Srv_Prepare() read 0 char"<<endl;
					continue; 
				}
				else
				{ //reading ok
					//INFO_STREAM << "plcsrvthread::run() in else cnt=" << cnt <<endl;
					header_t *head;
					head = reinterpret_cast<header_t *>(&buf[0]);
					uint16_t btype = head->db_number;
					btype = ntohs(btype);
					size_t ui_size = sizeof(data_t);//TODO
					INFO_STREAM << "plcsrvthread::run() before while cnt=" << cnt << " ui_size="<<ui_size <<endl;
					while(cnt < ui_size && cnt >= 0)
					{
						if(cnt > 0)
							DEBUG_STREAM << "plcsrvthread::run() after while cnt=" << cnt << " ui_size="<<ui_size <<endl;
						if(cnt < ui_size)
						{
							try
							{
								Tango::DevLong count=t_PLCRack3->input_length();
								//INFO_STREAM << "plcsrvthread::run() in while count=" << count <<endl;
								if(count > 0)
								{
									if(BUF_LEN-cnt < ui_size-cnt)
									{
										INFO_STREAM << "plcsrvthread::run() bytes missing=" << ui_size-cnt << " cannot fit in buffer, " << BUF_LEN-cnt << " bytes left!" <<endl;
										//empty buffer and restart reading
										buf = save_buf_ptr;
										memset (buf,0,2*BUF_LEN);
										cnt = 0;
									}
									//read just what's missing
									if(count > ui_size-cnt)
										count = ui_size-cnt;
									if(count > BUF_LEN-cnt)
										count = BUF_LEN-cnt;
									vector<unsigned char> dtr;
									t_PLCRack3->read(count, dtr);
									for (unsigned int i=0; i<dtr.size(); ++i)
									{
										buf[cnt+i] = dtr[i];
									}
									gettimeofday(&last_read, NULL);
									head = reinterpret_cast<header_t *>(&buf[0]);
									btype = head->db_number;
									btype = ntohs(btype);
								}
								else
								{
									timespec time_to_sleep;
									time_to_sleep.tv_sec=0;
									//time_to_sleep.tv_nsec= 0.5 * 1000000L;//0.5ms
									//time_to_sleep.tv_nsec= 10 * 1000000L;//100ms
									time_to_sleep.tv_nsec= 100 * 1000000L;//100ms//TODO
									nanosleep(&time_to_sleep, NULL);
								}
								cnt += count;
								if(count > 0)
									INFO_STREAM << "plcsrvthread::run() receiving for type:" << btype <<" "<< count << " more bytes, total received="<<cnt<< " expected="<<ui_size <<endl;
							}
							catch(Tango::DevFailed &e)
							{
								INFO_STREAM << __func__ << ": error reading " << e.errors[0].desc;
								cnt = -1;
								sleep(1);
							}
						}

						if(cnt < ui_size)
						{
							if(cnt > 0)
								INFO_STREAM << "plcsrvthread::run() still missing bytes total received=" << cnt <<" < expected="<< ui_size << endl;
							continue;
						}

						INFO_STREAM << "plcsrvthread::run() finally received complete type:" << btype <<" length:"<< ui_size<<endl;

					}//end while(cnt < ui_size)
					while(cnt > 0)
					{
						head = reinterpret_cast<header_t *>(&buf[0]);
						btype = head->db_number;
						btype = ntohs(btype);
						size_t ui_size = sizeof(data_t);//TODO
						DEBUG_STREAM << "plcsrvthread::run() read " << cnt << " in buffer, type:" << btype <<endl;
						if(cnt >= ui_size && ui_size > 0)
						{
							parse_message(btype, ui_size);
							cnt -= ui_size;
							if(cnt > 0)
							{
								buf += ui_size;
								INFO_STREAM << "plcsrvthread::run() after parsing " << ui_size << " there are " << cnt <<" remaining in buffer. buf moved ahead of "<< ui_size << " -> " << hex<<setw(2)<<setfill(' ')<< (int)(unsigned char) buf[0]<<","<<(int)(unsigned char) buf[1]<<","<< (int)(unsigned char) buf[2]<<","<<(int)(unsigned char) buf[3]<<","<< (int)(unsigned char) buf[4]<<","<<(int)(unsigned char) buf[5]<<","<<(int)(unsigned char) buf[6]<<","<< (int)(unsigned char) buf[7]<<endl;
							}
							else
							{
								buf = save_buf_ptr;
								memset (buf,0,2*BUF_LEN);
								INFO_STREAM << "plcsrvthread::run() buf empty after parsing cnt=" << cnt;
								break;
							}
						}
						else if(cnt > 0 && ui_size == 0 && btype == 0)
						{
							buf = save_buf_ptr;
							memset (buf,0,2*BUF_LEN);
							INFO_STREAM << "plcsrvthread::run() emptying buffer cnt=" << cnt << " type==0 && ui_size==0";
							cnt = 0;
							break;
						}
						else if(cnt < ui_size)
						{
							INFO_STREAM << __func__ << " exiting parsing loop with " << cnt << " remaining, type=" << btype << " size="<<ui_size;
							char tmp[BUF_LEN];
							memcpy(tmp,buf,cnt);
							buf = save_buf_ptr;
							memset (buf,0,2*BUF_LEN);
							memcpy(buf,tmp,cnt);						
							break;
						}
					}
					DEBUG_STREAM << "plcsrvthread::run() end of loop while (cnt > 0)"<<endl;

				}  // end of message-print else 
			// end of do loop statement 
			}while ((cnt >= 0)&&(!t_PLCRack3->abortflag));  // do loop condition 
		}//end of accept else
	}  // end of while (1) 
//	close(t_PLCRack3->sock); 
	server_started=0;
	return;
	
}


void plcsrvthread::parse_message(uint16_t type, uint16_t size)
{
	INFO_STREAM << __func__ << ": entering parsing, type="<<type<< " size=" <<size << endl;
	for(int k=0;k<size;)
	{
		stringstream tmp;
		int g;
		tmp << "DB"<<type<<": "<<setw(3)<<setfill(' ')<<k<<" -> ";
		for(g=0; g<10; g++)
		{
			if(k+g>=size)
				break;
			tmp << " " <<hex<<setw(2)<<setfill(' ')<< (int)(unsigned char)buf[k+g];
		}
		k+=g;
		if(k<size)
		{
			tmp << " |  "<<dec<<setw(3)<<setfill(' ')<<k<<hex<<" -> ";
			for(g=0; g<10; g++)
			{
				if(k+g>=size)
					break;
				tmp << " " <<hex<<setw(2)<<setfill(' ')<< (int)(unsigned char)buf[k+g];
			}
			k+=g;
		}
		DEBUG_STREAM<<tmp.str()<<dec<<endl;
	}
	switch(type)
	{
		case 500:
		case 0:
		case 200:
		case 201:
		case 202:
		case 210:
		case 211:
		case 212:
			INFO_STREAM << __func__ << ": received DB500 size=" <<size << endl;
			memcpy(&t_PLCRack3->db500, buf, size);
		break;
		default:
			INFO_STREAM << __func__ << ": message NOT recognized, type "<<type<< " size=" <<size << endl;
		break;
	}
}

uint16_t plcsrvthread::plc_swap(uint16_t in)
{
	return ((in>>8)|((in&0x00ff)<<8));
}



uint32_t plcsrvthread::plc_swaplong(uint32_t in)
{
	return ((in>>16)|((in&0x0000ffff)<<16));
}

float plcsrvthread::htonf(float val_f)
{
	float *ret_val;
	uint32_t *val_l;
	uint32_t ret_val_l;
	
	val_l = (uint32_t *)&val_f;
#if 1
	ret_val_l = htonl(*val_l);
#else
# if __BYTE_ORDER == __LITTLE_ENDIAN
	ret_val_l = __bswap_32(*val_l);
#else
	ret_val_l = *val_l;
#endif
#endif
	ret_val=(float *)&ret_val_l;

	return *ret_val;
}

float plcsrvthread::ntohf(float val_f)
{
	float *ret_val;
	uint32_t *val_l;
	uint32_t ret_val_l;
	
	val_l = (uint32_t *)&val_f;
#if 1
	ret_val_l = ntohl(*val_l);
#else
# if __BYTE_ORDER == __LITTLE_ENDIAN
	ret_val_l = __bswap_32(*val_l);
#else
	ret_val_l = *val_l;
#endif
#endif
	ret_val=(float *)&ret_val_l;

	return *ret_val;
}

}//namespace


